#! /usr/bin/perl -w
use FindBin;
use lib "$FindBin::Bin";
use BitTorrent::BDecode;
use Digest::SHA1 qw(sha1);
use LWP::UserAgent;


# TorrentSniff  http://www.highprogrammer.com/alan/perl/torrentsniff.html
# - 2003-05-29 - Changes by Alan De Smet (http://www.highprogrammer.com/alan/)
#      - Alan's changes are released under the "MIT License" as listed below:
#        (And frankly, they're pretty dumb changes, I mostly just slashed and
#        burned stuff I didn't need or couldn't implement from TorrentSpy.)
#
# - Based on vesion 0.1.0.3-BETA of TorrentSpy from http://torrentspy.sf.net/
#   Copyright 2003 "knowbuddy" who is at users.sourceforge.net
#   SourceForge indicates "MIT License"
#
#   The MIT License:
#
#	Permission is hereby granted, free of charge, to any person obtaining a
#	copy of this software and associated documentation files (the "Software"),
#	to deal in the Software without restriction, including without limitation
#	the rights to use, copy, modify, merge, publish, distribute, sublicense,
#	and/or sell copies of the Software, and to permit persons to whom the
#	Software is furnished to do so, subject to the following conditions:
#
#	The above copyright notice and this permission notice shall be included in
#	all copies or substantial portions of the Software.
#
#	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#	DEALINGS IN THE SOFTWARE.

use strict;

my $VERSION = '0.1';
my $PROGRAM_NAME = 'TorrentSniff';


main();

exit 0;

sub main {
	if(not @ARGV) {
		usage_exit();
	}
	foreach my $arg (@ARGV) {
		if($arg =~ m|://|) {
			process_url($arg);
		} else {
			process_file($arg);
		}
		print "\n";
	}
}

sub process_url {
	my($url) = @_;
	print "$url\n";
	my $res = get_url($url);
	if(not defined $res->is_success) {
		print STDERR "Problem downloading $url:\n";
		print STDERR $res->status_line;
		return;
	}
	output_torrent_data($res->content);
}

sub process_file {
	my($file) = @_;
	print "$file\n";
	local *TOR;
	if( not open(TOR, "<", $file) ) {
		print STDERR "Unable read $file because of $!\n";
		return 0;
	}
	binmode(TOR);
	my $body;
	read(TOR, $body, (-s $file));
	close(TOR);
	output_torrent_data($body);
}

sub output_torrent_data {
	my($body) = @_;
	my $result = process_torrent_data($body);
	if(not defined $result) {
		print STDERR "Problem reading torrent file\n";
	}

	my $scrape_url = get_tracker_status_url($result->{'announce'});
	my $res = get_url($scrape_url);
	if(not defined $res->is_success) {
		print STDERR "Problem contacting tracker at $scrape_url:\n";
		print STDERR $res->status_line;
		return undef;
	}
	my $tracker_status = get_tracker_status($res->content_ref);

	print  "   info hash:      $result->{'hash'}\n";
	print  "   announce url:   $result->{'announce'}\n";

	if(defined $tracker_status) {
		my $hash = $result->{'hash'};
		my %status = %{$tracker_status};
		if(exists $status{$hash}) {
			my $seeds = $status{$hash}->{'complete'};
			my $leeches = $status{$hash}->{'incomplete'};
			print "   full copies:    $seeds seeds\n";
			print "   partial copies: $leeches leeches\n";
		} else {
			print STDERR "    Cannot find hash $hash on tracker.\n";
			print STDERR "    Torrent may have been removed from tracker.\n";
		}
	} else {
		print "   Problem contacting tracker\n";
	}

	printf "   %15s %s\n", 'Bytes', 'File';
	foreach my $file (@{$result->{'files'}}) {
		my $size = $file->{'size'};
		my $name = $file->{'name'};
		printf "   %15s %s\n", commify($size), $name;
	}
	if(@{$result->{'files'}} > 1) {
		printf "   %15s %s\n", commify($result->{'total_size'}), 'TOTAL';
	}
}

sub process_torrent_data {
	my($body) = @_;

	my %result;

	my $t = BitTorrent::BDecode::bdecode(\$body);

	my $info = $t->{'info'};
	my $s = substr($body, $t->{'_info_start'}, $t->{'_info_length'});
	my $hash = bin2hex(sha1($s));
	my $announce = $t->{'announce'};

	$result{'hash'} = $hash;
	$result{'announce'} = $announce;
	$result{'files'} = [];
	my $tsize = 0;
	if(defined($info->{'files'})) {
		foreach my $f (@{$info->{'files'}}) {
			my %file_record = ( 'size' => $f->{'length'});

			$tsize += $f->{'length'};
			my $path = $f->{'path'};

			if(ref($path) eq 'ARRAY') {
				$file_record{'name'} = $info->{'name'}.'/'.$path->[0];
			} else {
				$file_record{'name'} = $info->{'name'}.'/'.$path;
			}
			push @{$result{'files'}}, \%file_record;

		}
	} else {
		$tsize += $info->{'length'},

		push @{$result{'files'}}, 
			{
				'size' => $info->{'length'},
				'name' => $info->{'name'},
			};

	}
	$result{'total_size'} = $tsize;

	return \%result;
}

# Given a tracker announce URL, return the status ("scrape") URL
sub get_tracker_status_url {
	my($url) = @_;
	$url =~ s|/announce$|/scrape|ig;
	return $url;
}

# Retrieve complete (seed) and incomplete (leech) counts for all
# torrents at a particular tracker.
sub get_tracker_status {
	my($status_body) = @_;

	my $s;
	eval { $s = BitTorrent::BDecode::bdecode($status_body); };
	if($@) {
		print STDERR "Invalid tracker response $@";
		return undef;
	}

	if(not exists $s->{'files'}) {
		print STDERR "Tracker returned odd results (no files)\n";
		return undef;
	}

	my %results;
	foreach my $f (%{$s->{'files'}}) {
		my $v = $s->{'files'}{$f};
		my $fhash = bin2hex($f);
		my $seeds = $v->{'complete'} || "0";
		my $leeches = $v->{'incomplete'} || "0";
		if(exists $results{$fhash}) {
			print STDERR "Tracker has hash $fhash multiple times\n";
		}
		$results{$fhash} = {'complete' => $seeds, 'incomplete' => $leeches};
	}
	return \%results;
}

# Add commas to number
sub commify {
  local $_ = shift;
  1 while s/^([-+]?\d+)(\d{3})/$1,$2/;
  return $_;
}

# Encode binary as hex characters
sub bin2hex {
  my ($d) = @_;
  $d =~ s/(.)/sprintf("%02x",ord($1))/egs;
  $d = lc($d);
  return $d;
}

BEGIN {
my $ua;
sub get_url {
	if(not defined $ua) {
		$ua = LWP::UserAgent->new(
			'env_proxy' => 1,
			'agent'		 => "$PROGRAM_NAME/$VERSION ",
			'timeout'	 => 15,
		);
	}
	return $ua->get($_[0]);
}
}

sub usage_exit {
	print <<END;
$0: Call with one or more .torrent files or URLs to .torrent files.
END
}

